from typing import Generic, Optional, Set, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
	from .Node import Node

TData = TypeVar("TData")
TNode = TypeVar("TNode", bound="Node")


class Edge(Generic[TNode, TData]):
	"""Represents a way to reach one node from another.

	Args:
		Generic (TNode): The type of the nodes of the graph.
		Generic (TData): Data to be contained by an edge. For example if the edge represents getting from one chess board position to another, the data might be "Be5".
	"""

	def __init__(
		self,
		node_a: TNode,
		node_b: TNode,
		data: Optional[TData] = None,
		weight: int = 1,
	):
		"""Construct a node.

		Args:
			node_a (TNode): One of the endpoints of the edge
			node_b (TNode): The other endpoint of the edge.
			data (TData, optional): Data to be stored by the node. Defaults to None.
			weight (int): Weight of the edge. Defaults to 1.
		"""
		self.__node_a = node_a
		self.__node_b = node_b
		self.__data = data
		self.__weight = weight

	@property
	def node_a(self) -> TNode:
		"""The first node of the edge."""
		return self.__node_a

	@property
	def node_b(self) -> TNode:
		"""The second node of the edge."""
		return self.__node_b

	@property
	def data(self) -> Optional[TData]:
		"""The data stored in this edge."""
		return self.__data

	@property
	def weight(self) -> int:
		return self.__weight

	@property
	def __nodes_set(self) -> Set[TNode]:
		"""A set of the two nodes in this edge."""
		return {self.node_a, self.node_b}

	def __eq__(self, other):
		return isinstance(other, Edge) and self.__nodes_set == other.__nodes_set

	def __hash__(self) -> int:
		return hash(self.node_a) + hash(self.node_b)

	def __str__(self) -> str:
		return str(self.data) if self.data is not None else str(self.__nodes_set)

	def __repr__(self) -> str:
		return str(self)


from enum import Enum
from .Frontier import Frontier
from .IterativeDeepeningFrontier import IterativeDeepeningFrontier
from .DepthFirstFrontier import DepthFirstFrontier
from .BreadthFirstFrontier import BreadthFirstFrontier
from .Node import Node, TNode
from .Edge import TData

frontiers = {
	"BreadthFirstSearch": lambda start_node: BreadthFirstFrontier[TNode, TData](
		start_node
	),
	"DepthFirstSearch": lambda start_node: DepthFirstFrontier[TNode, TData](start_node),
	"IterativeDeepeningSearch": lambda start_node: IterativeDeepeningFrontier[
		TNode, TData
	](start_node),
}


class Algorithm(Enum):
	DepthFirstSearch = 1
	BreadthFirstSearch = 2
	IterativeDeepeningSearch = 3

	def new_frontier(self, start: Node[TNode, TData]) -> Frontier[TNode, TData]:
		"""Construct and return a frontier for this algorithm.

		Args:
			Generic (TNode): The type of the nodes of the graph.
			Generic (TData): The type of the data stored by the edges of the graph.
			start (Node[TNode, TData] The start node for the search.

		Returns:
			Frontier[TNode, TData]: A frontier for the search capable of containing nodes of type TNode with edges capable of containing data of type TData.
		"""
		return frontiers[self.name](start)


from typing import Generic, Sequence, List, Union, overload
from .Node import TNode
from .Edge import Edge, TData


class Path(Generic[TNode, TData], Sequence[Edge[TNode, TData]]):
	"""A sequence of edges from one node to another.

	Args:
		Generic (TNode): The type of the nodes of the graph.
		Generic (TData): The type of the data stored by the edges of the graph.
	"""

	def __init__(self, edges: Sequence[Edge[TNode, TData]] = ()):
		self.__edges = list(edges)

	def weight(self) -> int:
		return sum(edge.weight for edge in self.__edges)

	@property
	def edges(self) -> List[Edge[TNode, TData]]:
		return self.__edges.copy()

	@property
	def nodes(self) -> List[TNode]:
		return (
			[self.__edges[0].node_a] + [edge.node_b for edge in self.__edges]
			if self.__edges
			else []
		)

	def __len__(self) -> int:
		return len(self.__edges)

	def __bool__(self) -> bool:
		return bool(self.__edges)

	@overload
	def __getitem__(self, idx: int) -> Edge[TNode, TData]:
		...

	@overload
	def __getitem__(self, s: slice) -> Sequence[Edge[TNode, TData]]:
		...

	def __getitem__(
		self, idx: Union[int, slice]
	) -> Union[Edge[TNode, TData], Sequence[Edge[TNode, TData]]]:
		return self.__edges[idx]

	def __add__(self, other: Sequence[Edge[TNode, TData]]) -> "Path[TNode, TData]":
		return Path(self.__edges + list(other))


from abc import ABC, abstractmethod
from typing import Generic, TYPE_CHECKING, Iterable, TypeVar
from .Edge import TNode, TData, Edge


class Node(Generic[TNode, TData], ABC):
	"""An abstract base class for a node in a graph.

	Args:
		Generic (TNode): The concrete type of the nodes in the graph.
		Generic (TData): The type of the data stored by the edges of the graph.
	"""

	@abstractmethod
	def neighbours(self) -> Iterable[Edge[TNode, TData]]:
		"""Get the edges to all the nodes adjacent to this one.

		Returns:
			Iterable[Edge[TNode, TData]]: The edges to all the nodes adjacent to this one.
		"""
		pass



from typing import Generic, Optional, Set
from .Heap import Heap
from .Frontier import Frontier
from .State import State
from .Node import TNode
from .Edge import TData


class DijkstraFrontier(Generic[TNode, TData], Frontier[TNode, TData]):
	"""A frontier for a graph search algorithm that performs Dijkstra's algorithm.

	Args:
		Generic (TNode): The type of the nodes of the graph.
		Generic (TData): The type of the data stored by the edges of the graph.
	"""

	def __init__(self, start: TNode):
		self.__heap = Heap[State[TNode, TData]](key=lambda state: state.path.weight())
		self.__heap.push(State(start))
		self.__visited: Set[TNode] = set()

	def extract(self) -> Optional[State[TNode, TData]]:
		self.__remove_visited()
		state = self.__heap.pop()
		self.__visited.add(state.node)
		return state

	def insert(self, state: State[TNode, TData]):
		self.__heap.push(state)

	def __remove_visited(self):
		"""Remove visited nodes from the heap until the top element is unvisited."""
		while self.__heap and self.__heap.peek() in self.__visited:
			self.__heap.pop()

	def __len__(self) -> int:
		"""Get the number of elements in the frontier.

		Note:
			This number may include elements that have already been popped, in which case they'll not be popped again. Therefore this is really an upper bound on the number of elements that can be popped. The only other thing guaranteed is that when there are no unvisited elements, this function will return 0 (as required by the contract of Frontier).

		Returns:
			int: The number of elements in the frontier.
		"""
		self.__remove_visited()
		return len(self.__heap)


from abc import ABC, abstractmethod
from typing import Generic, Optional, Sized
from .State import State
from .Node import TNode
from .Edge import TData


class Frontier(ABC, Sized, Generic[TNode, TData]):
	"""A data structure which holds the furthest nodes we have explored and it determines the order in which we will explore them further.

	Args:
		Generic (TNode): The type of the nodes of the graph.
		Generic (TData): The type of the data stored by the edges of the graph.

	To implement your own frontier, you must subclass this class and implement the following methods:
	- `extract(self) -> Optional[State]` to extract the next state from the frontier.
	- `insert(self, state: State) -> Optional[State]` to insert a state into the frontier.
	"""

	def is_empty(self) -> bool:
		"""Check if the frontier has more states to extract.

		Returns:
			bool: True if the frontier is empty, false otherwise.
		"""
		return len(self) == 0

	@abstractmethod
	def extract(self) -> Optional[State[TNode, TData]]:
		"""Extract a state to expand from the frontier.

		Returns:
			State: The state to expand, or None if the frontier is empty.
		"""
		pass

	@abstractmethod
	def insert(self, state: State[TNode, TData]):
		"""Insert a state into the frontier.

		Args:
			state (State): The state to insert.
		"""
		pass

	def __bool__(self) -> bool:
		return not self.is_empty()


from typing import Any, Callable, Generic, Iterable, List, Tuple, TypeVar
import heapq
import deprecation

T = TypeVar("T")


class Heap(Generic[T]):
	"""A heap (aka priority queue) data structure.

	Args:
		Generic (T): The type of the elements in the heap.
	"""

	def __init__(
		self, items: Iterable[T] = None, key: Callable[[T], Any] = lambda _: _
	):
		"""Construct a heap data structure.

		Args:
			items (Iterable[T], optional): The items to initialize the heap with. If omitted will create an empty heap.
			key (Callable[[T], Any], optional): A function to be used to determine the priority of the elements in the heap. It should take an element of type T and return some comparable object (e.g. an integer) which represents the priority. Lower priorities are extracted first. If no priority function is provided, then the inserted objects themselves are used and a min-heap is created.
		"""
		self.__heap_key = key
		self.__heap_index = -1
		self.__heap_list = [self.__heap_node(item) for item in (items or [])]
		heapq.heapify(self.__heap_list)

	def push(self, item: T, key: Any = None):
		"""Insert a new item into the heap.

		Args:
			item (T): The item to insert.
			key (Any, optional): Override the key returned by the `key` function provided in the constructor.
		"""
		heapq.heappush(self.__heap_list, self.__heap_node(item, key))

	def pop(self) -> T:
		"""Extrace an item from the heap.

		Returns:
			T: The item extracted from the heap.

		Raises:
			IndexError: If there are no more items in the heap.
		"""
		return heapq.heappop(self.__heap_list)[-1]

	@deprecation.deprecated("0.0.2", details="Use len(heap) instead.")
	def size(self) -> int:
		"""Get the number of items in the heap.

		Returns:
			int: The number of items in the heap.
		"""
		return len(self)

	def items(self) -> List[T]:
		"""Get the items in the heap.

		Returns:
			List[T]: The items in the heap.
		"""
		return [node[-1] for node in self.__heap_list]

	def show(self):
		"""Print the items in the heap."""
		print(*self.items())

	def peek(self) -> T:
		"""Get the first item from the heap without extracting it.

		Returns:
			T: The first item in the heap.

		Raises:
			IndexError: If the heap is empty.
		"""
		return self.__heap_list[0][-1]

	def __len__(self):
		return len(self.__heap_list)

	def __bool__(self):
		return bool(self.__heap_list)

	def __heap_node(self, item: T, key: Any = None) -> Tuple[Any, int, T]:
		"""Create a comparable tuple that can be inserted into the heap's list.

		Args:
			item (T): The item to be inserted.
			key (Any, optional): Override the key returned by the `key` function provided in the constructor.

		Returns:
			Tuple[Any, int, T]: The tuple that can be inserted into the heap's list.
		"""
		self.__heap_index += 1
		return (self.__heap_key(item) if key is None else key, self.__heap_index, item)


if __name__ == "__main__":
	print("Nothing to Show")



from typing import Generic, Optional, Set
from .State import State
from .Frontier import Frontier
from .Stack import Stack
from .Node import Node, TNode
from .Edge import TData


class DepthFirstFrontier(Generic[TNode, TData], Frontier[TNode, TData]):
	"""A frontier for a graph search algorithm that performs Depth First Search (DFS) using a stack.

	Args:
		Generic (TNode): The type of the nodes of the graph.
		Generic (TData): The type of the data stored by the edges of the graph.
	"""

	def __init__(self, start: TNode):
		self.__stack = Stack[State[TNode, TData]]()
		self.__stack.push(State(start))
		self.__visited: Set[TNode] = set()

	def extract(self) -> Optional[State[TNode, TData]]:
		return None if self.__stack.is_empty() else self.__stack.pop()

	def insert(self, state: State[TNode, TData]):
		if state.node not in self.__visited:
			self.__visited.add(state.node)
			self.__stack.push(state)

	def __len__(self) -> int:
		return len(self.__stack)


from typing import Iterable, Set
from pygraphsearch.Edge import Edge, TData, TNode
from pygraphsearch.Node import Node


class SimpleNode(Node["SimpleNode", str]):
	def __init__(self, name: str):
		self.__name = name
		self.__edges: Set[Edge["SimpleNode", str]] = set()

	def connect(self, node: "SimpleNode", weight: int, edge_name: str = ""):
		edge = Edge(self, node, edge_name, weight)
		self.__edges.add(edge)
		node.__edges.add(edge)

	@property
	def name(self) -> str:
		return self.__name

	def neighbours(self) -> Iterable[Edge["SimpleNode", str]]:
		return self.__edges.copy()

	def __eq__(self, other):
		return isinstance(other, self.__class__) and self.name == other.name

	def __hash__(self) -> int:
		return hash(self.name)


from .search import search
from .State import State
from .Node import Node, TNode
from .Edge import Edge, TData
from .Frontier import Frontier
from .IterativeDeepeningFrontier import IterativeDeepeningFrontier
from .BreadthFirstFrontier import BreadthFirstFrontier
from .DepthFirstFrontier import DepthFirstFrontier
from .DijkstraFrontier import DijkstraFrontier
from .Stack import Stack
from .Algorithm import Algorithm


from typing import Generic, List, TypeVar


T = TypeVar("T")


class Stack(Generic[T]):
	"""A collection of items which can be accessed in a LIFO order.

	Args:
		Generic (T): The type of the items stored in the stack.
	"""

	def __init__(self) -> None:
		"""Construct an empty stack."""
		super().__init__()
		self.__list: List[T] = []

	def is_empty(self) -> bool:
		"""Check if the stack is empty.

		Returns:
			bool: True if the stack is empty, false otherwise.
		"""
		return not bool(self.__list)

	def push(self, value: T):
		"""Push an item onto the stack.

		Args:
			value (T): The item to push onto the stack.
		"""
		self.__list.append(value)

	def pop(self) -> T:
		"""Pop an item from the top of the stack.

		Returns:
			T: The topmost item that was removed from the stack.

		Raises:
			IndexError: If the stack was empty.
		"""
		return self.__list.pop()

	def __len__(self) -> int:
		return len(self.__list)

	def __bool__(self) -> bool:
		return bool(self.__list)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
	long_description = fh.read()

setuptools.setup(
	name="pygraphsearch",
	version="1.3.1",
	author="Abraham Murciano",
	author_email="abrahammurciano@gmail.com",
	description="A python package to search graphs.",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/abrahammurciano/pygraphsearch",
	project_urls={
		"Bug Tracker": "https://github.com/abrahammurciano/pygraphsearch/issues",
	},
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	package_dir={"": "src"},
	package_data={"pygraphsearch": ["py.typed"]},
	packages=setuptools.find_packages(where="src"),
	python_requires=">=3.6",
	test_suite="nose.collector",
	tests_require=["nose"],
)



from typing import Generic, Optional, Set
from collections import deque
from .Frontier import Frontier
from .State import State
from .Node import TNode
from .Edge import TData


class BreadthFirstFrontier(Generic[TNode, TData], Frontier[TNode, TData]):
	"""A frontier for a graph search algorithm that performs Breadth First Search (BFS) using a queue.

	Args:
		Generic (TNode): The type of the nodes of the graph.
		Generic (TData): The type of the data stored by the edges of the graph.
	"""

	def __init__(self, start: TNode):
		self.__queue = deque[State[TNode, TData]]()
		self.__queue.append(State(start))
		self.__visited: Set[TNode] = set()

	def extract(self) -> Optional[State[TNode, TData]]:
		return self.__queue.popleft() if self.__queue else None

	def insert(self, state: State[TNode, TData]):
		if state.node not in self.__visited:
			self.__visited.add(state.node)
			self.__queue.append(state)

	def __len__(self) -> int:
		return len(self.__queue)


from typing import Generic
from .Stack import Stack
from .Frontier import Frontier
from .State import State
from .Node import Node, TNode
from .Edge import TData


class IterativeDeepeningFrontier(Generic[TNode, TData], Frontier[TNode, TData]):
	"""A frontier for a graph search algorithm that performs DFS up to a limited depth, then if no path is found, increases the depth and starts again.

	Args:
		Generic (TNode): The type of the nodes of the graph.
		Generic (TData): The type of the data stored by the edges of the graph.
	"""

	def __init__(self, start: TNode, initial_depth: int = 1, depth_step: int = 1):
		"""Construct a frontier for an interative deepening search algorithm.

		Args:
			start (TNode): The node from which to start the search.
			initial_depth (int, optional): The initial depth limit for the search. Defaults to 1.
			depth_step (int, optional): The amount to increase the depth by each time. Defaults to 1.
		"""
		self.__stack = Stack[State[TNode, TData]]()
		self.__start_state = State[TNode, TData](start)
		self.__stack.push(self.__start_state)
		self.__depth = initial_depth
		self.__can_go_deeper = False
		self.__depth_step = depth_step

	def is_empty(self) -> bool:
		return self.__stack.is_empty() and not self.__can_go_deeper

	def insert(self, state: State[TNode, TData]):
		if len(state.path) <= self.__depth:
			self.__stack.push(state)
			self.__start_state = self.__start_state or state
		else:
			# ignore new state this time because this branch is too deep. But set indicator that there are branches that go deeper than the current depth.
			self.__can_go_deeper = True

	def extract(self):
		if self.__stack.is_empty():
			if self.__can_go_deeper:
				# start over with increased depth
				self.__depth += self.__depth_step
				self.__can_go_deeper = False
				return self.__start_state
			else:
				return None  # There is no path to the target
		return self.__stack.pop()

	def __len__(self) -> int:
		"""Get the number of states in the frontier.

		Returns:
			int: The number of states in the frontier.
		"""
		return len(self.__stack)


from unittest import TestCase

from .SimpleNode import SimpleNode
from pygraphsearch import search, DijkstraFrontier


class DijkstraTest(TestCase):
	def setUp(self):
		nodes = [SimpleNode(str(i)) for i in range(6)]
		nodes[0].connect(nodes[5], 5)
		nodes[0].connect(nodes[1], 1)
		nodes[1].connect(nodes[5], 3)
		nodes[1].connect(nodes[2], 1)
		nodes[2].connect(nodes[3], 1)
		nodes[2].connect(nodes[5], 1)
		nodes[3].connect(nodes[4], 1)
		nodes[4].connect(nodes[5], 1)
		self.nodes = nodes

	def test_dijkstra(self):
		dijkstra_frontier = DijkstraFrontier[SimpleNode, str](self.nodes[0])
		state = search(dijkstra_frontier, lambda node: node == self.nodes[5])
		self.assertEqual(state.node, self.nodes[5])
		self.assertEqual(
			state.path.nodes,
			[self.nodes[0], self.nodes[1], self.nodes[2], self.nodes[5]],
		)



from typing import Generic, Iterable, List, Optional

from .Path import Path
from .Edge import Edge
from .Node import TNode
from .Edge import TData


class State(Generic[TNode, TData]):
	"""Represents the state of a search.

	A state is made up of the current node and the path taken to get there.

	Args:
		Generic (TNode): The concrete type of the node that this state contains.
		Generic (TData): The type of the data stored in the edges.
	"""

	def __init__(self, node: TNode, path: Path = Path()):
		"""Construct a state.

		Args:
			node (TNode): The current node.
			path (Path, optional): The sequence of edges taken to reach the current node. Defaults to an empty path.
		"""
		self.__node = node
		self.__path = Path(path or [])

	@property
	def node(self) -> TNode:
		"""The current node in the search."""
		return self.__node

	@property
	def path(self) -> Path[TNode, TData]:
		"""The list of edges taken to reach the current node."""
		return self.__path

	def next_states(self) -> Iterable["State[TNode,TData]"]:
		"""The states reachable by taking one edge from the current node of this state. This does not include backtracking to the node before the current one.

		Returns:
			Iterable[State]: The states that can be reached from the current node by taking a single edge.
		"""
		return [
			State[TNode, TData](edge.node_b, self.__path + [edge])
			for edge in self.node.neighbours()
			if not self.path or edge != self.path[-1]
		]



from typing import Callable, Optional, Union, overload
from .Node import TNode
from .Edge import TData
from .Algorithm import Algorithm
from .State import State
from .Frontier import Frontier
from .Node import Node


@overload
def search(
	frontier: Frontier[TNode, TData], is_target: Callable[[TNode], bool], /
) -> Optional[State[TNode, TData]]:
	"""Search for the path to a target.

	Args:
		frontier (Frontier): An object derived from Frontier. It should contain only the start state and will be modified by this function.
		is_target (Callable[[TNode], bool]): A predicate which tells us if a node is a target.

	Returns:
		Optional[State]: The target state, or None if none exists.
	"""
	...


@overload
def search(
	start: Node[TNode, TData],
	is_target: Callable[[TNode], bool],
	algorithm: Algorithm,
	/,
) -> Optional[State[TNode, TData]]:
	"""Search for the path to a target

	Args:
		start (Node[TNode, TData] The node to start the search from.
		is_target (Callable[[TNode], bool]): A predicate which tells us if a node is a target.
		algorithm (Algorithm): The algorithm you want to use for the search.

	Returns:
		Optional[State]: The target state, or None if none exists.
	"""
	...


def search(
	arg1: Union[Frontier[TNode, TData], Node[TNode, TData]],
	arg2: Callable[[TNode], bool],
	arg3: Optional[Algorithm] = None,
	/,
) -> Optional[State[TNode, TData]]:
	is_target = arg2
	if isinstance(arg1, Frontier):
		frontier = arg1
		while not frontier.is_empty():
			state = frontier.extract()
			if state is None or is_target(state.node):
				return state
			for next_state in state.next_states():
				frontier.insert(next_state)
		return None

	elif isinstance(arg1, Node):
		start = arg1
		algorithm = arg3
		assert algorithm is not None
		return search(algorithm.new_frontier(start), is_target)

	raise TypeError("The first argument must be an instance of Node or Frontier")